@@ -1,9 +1,9 @@
 # Copyright (c) 2008 Giorgos Verigakis <verigak@gmail.com>
-# 
+#
 # Permission to use, copy, modify, and distribute this software for any
 # purpose with or without fee is hereby granted, provided that the above
 # copyright notice and this permission notice appear in all copies.
-# 
+#
 # THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 # WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 # MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
@@ -18,11 +18,34 @@ from ctypes import *
 from ctypes.util import find_library
 from errno import *
 from functools import partial
+from os import strerror
 from platform import machine, system
+from signal import signal, SIGINT, SIG_DFL
 from stat import S_IFDIR
 from traceback import print_exc
 
 
+__version__ = '1.1'
+
+_system = system()
+_machine = machine()
+
+if _system == 'Darwin':
+    _libfuse_path = find_library('fuse4x') or find_library('osxfuse') or \
+                    find_library('fuse')
+else:
+    _libfuse_path = find_library('fuse')
+if not _libfuse_path:
+    raise EnvironmentError('Unable to find libfuse')
+
+if _system == 'Darwin':
+    _libiconv = CDLL(find_library('iconv'), RTLD_GLOBAL) # libfuse dependency
+_libfuse = CDLL(_libfuse_path)
+
+if _system == 'Darwin' and hasattr(_libfuse, 'macfuse_version'):
+    _system = 'Darwin-MacFuse'
+
+
 class c_timespec(Structure):
     _fields_ = [('tv_sec', c_long), ('tv_nsec', c_long)]
 
@@ -32,9 +55,8 @@ class c_utimbuf(Structure):
 class c_stat(Structure):
     pass    # Platform dependent
 
-_system = system()
-if _system in ('Darwin', 'FreeBSD'):
-    _libiconv = CDLL(find_library("iconv"), RTLD_GLOBAL)     # libfuse dependency
+
+if _system in ('Darwin', 'Darwin-MacFuse', 'FreeBSD'):
     ENOTSUP = 45
     c_dev_t = c_int32
     c_fsblkcnt_t = c_ulong
@@ -48,20 +70,41 @@ if _system in ('Darwin', 'FreeBSD'):
         c_size_t, c_int, c_uint32)
     getxattr_t = CFUNCTYPE(c_int, c_char_p, c_char_p, POINTER(c_byte),
         c_size_t, c_uint32)
-    c_stat._fields_ = [
-        ('st_dev', c_dev_t),
-        ('st_ino', c_uint32),
-        ('st_mode', c_mode_t),
-        ('st_nlink', c_uint16),
-        ('st_uid', c_uid_t),
-        ('st_gid', c_gid_t),
-        ('st_rdev', c_dev_t),
-        ('st_atimespec', c_timespec),
-        ('st_mtimespec', c_timespec),
-        ('st_ctimespec', c_timespec),
-        ('st_size', c_off_t),
-        ('st_blocks', c_int64),
-        ('st_blksize', c_int32)]
+    if _system == 'Darwin':
+        c_stat._fields_ = [
+            ('st_dev', c_dev_t),
+            ('st_mode', c_mode_t),
+            ('st_nlink', c_uint16),
+            ('st_ino', c_uint64),
+            ('st_uid', c_uid_t),
+            ('st_gid', c_gid_t),
+            ('st_rdev', c_dev_t),
+            ('st_atimespec', c_timespec),
+            ('st_mtimespec', c_timespec),
+            ('st_ctimespec', c_timespec),
+            ('st_birthtimespec', c_timespec),
+            ('st_size', c_off_t),
+            ('st_blocks', c_int64),
+            ('st_blksize', c_int32),
+            ('st_flags', c_int32),
+            ('st_gen', c_int32),
+            ('st_lspare', c_int32),
+            ('st_qspare', c_int64)]
+    else:
+        c_stat._fields_ = [
+            ('st_dev', c_dev_t),
+            ('st_ino', c_uint32),
+            ('st_mode', c_mode_t),
+            ('st_nlink', c_uint16),
+            ('st_uid', c_uid_t),
+            ('st_gid', c_gid_t),
+            ('st_rdev', c_dev_t),
+            ('st_atimespec', c_timespec),
+            ('st_mtimespec', c_timespec),
+            ('st_ctimespec', c_timespec),
+            ('st_size', c_off_t),
+            ('st_blocks', c_int64),
+            ('st_blksize', c_int32)]
 elif _system == 'Linux':
     ENOTSUP = 95
     c_dev_t = c_ulonglong
@@ -75,7 +118,6 @@ elif _system == 'Linux':
     setxattr_t = CFUNCTYPE(c_int, c_char_p, c_char_p, POINTER(c_byte), c_size_t, c_int)
     getxattr_t = CFUNCTYPE(c_int, c_char_p, c_char_p, POINTER(c_byte), c_size_t)
     
-    _machine = machine()
     if _machine == 'x86_64':
         c_stat._fields_ = [
             ('st_dev', c_dev_t),
@@ -179,6 +221,8 @@ class fuse_context(Structure):
         ('pid', c_pid_t),
         ('private_data', c_voidp)]
 
+_libfuse.fuse_get_context.restype = POINTER(fuse_context)
+
 class fuse_operations(Structure):
     _fields_ = [
         ('getattr', CFUNCTYPE(c_int, c_char_p, POINTER(c_stat))),
@@ -238,13 +282,6 @@ def set_st_attrs(st, attrs):
             setattr(st, key, val)
 
 
-_libfuse_path = find_library('fuse')
-if not _libfuse_path:
-    raise EnvironmentError('Unable to find libfuse')
-_libfuse = CDLL(_libfuse_path)
-_libfuse.fuse_get_context.restype = POINTER(fuse_context)
-
-
 def fuse_get_context():
     """Returns a (uid, gid, pid) tuple"""
     ctxp = _libfuse.fuse_get_context()
@@ -252,6 +289,11 @@ def fuse_get_context():
     return ctx.uid, ctx.gid, ctx.pid
 
 
+class FuseOSError(OSError):
+    def __init__(self, errno):
+        super(FuseOSError, self).__init__(errno, strerror(errno))
+
+
 class FUSE(object):
     """This class is the lower level interface and should not be subclassed
        under normal use. Its methods are called by fuse.
@@ -283,9 +325,17 @@ class FUSE(object):
             if prototype != c_voidp and getattr(operations, name, None):
                 op = partial(self._wrapper_, getattr(self, name))
                 setattr(fuse_ops, name, prototype(op))
-        _libfuse.fuse_main_real(len(args), argv, pointer(fuse_ops),
-            sizeof(fuse_ops), None)
+        
+        old_handler = signal(SIGINT, SIG_DFL)
+        
+        err = _libfuse.fuse_main_real(len(args), argv, pointer(fuse_ops),
+                                      sizeof(fuse_ops), None)
+        
+        signal(SIGINT, old_handler)
+        
         del self.operations     # Invoke the destructor
+        if err:
+            raise RuntimeError(err)
     
     def _wrapper_(self, func, *args, **kwargs):
         """Decorator for the methods that follow"""
@@ -331,6 +381,11 @@ class FUSE(object):
         return self.operations('chmod', path, mode)
     
     def chown(self, path, uid, gid):
+        # Check if any of the arguments is a -1 that has overflowed
+        if c_uid_t(uid + 1).value == 0:
+            uid = -1
+        if c_gid_t(gid + 1).value == 0:
+            gid = -1
         return self.operations('chown', path, uid, gid)
     
     def truncate(self, path, length):
@@ -482,34 +537,34 @@ class FUSE(object):
 
 class Operations(object):
     """This class should be subclassed and passed as an argument to FUSE on
-       initialization. All operations should raise an OSError exception on
-       error.
+       initialization. All operations should raise a FuseOSError exception
+       on error.
        
        When in doubt of what an operation should do, check the FUSE header
        file or the corresponding system call man page."""
     
     def __call__(self, op, *args):
         if not hasattr(self, op):
-            raise OSError(EFAULT, '')
+            raise FuseOSError(EFAULT)
         return getattr(self, op)(*args)
-        
+    
     def access(self, path, amode):
         return 0
     
     bmap = None
     
     def chmod(self, path, mode):
-        raise OSError(EROFS, '')
+        raise FuseOSError(EROFS)
     
     def chown(self, path, uid, gid):
-        raise OSError(EROFS, '')
+        raise FuseOSError(EROFS)
     
     def create(self, path, mode, fi=None):
         """When raw_fi is False (default case), fi is None and create should
            return a numerical file handle.
            When raw_fi is True the file handle should be set directly by create
            and return 0."""
-        raise OSError(EROFS, '')
+        raise FuseOSError(EROFS)
     
     def destroy(self, path):
         """Called on filesystem destruction. Path is always /"""
@@ -533,11 +588,11 @@ class Operations(object):
            while Linux counts only the subdirectories."""
         
         if path != '/':
-            raise OSError(ENOENT, '')
+            raise FuseOSError(ENOENT)
         return dict(st_mode=(S_IFDIR | 0755), st_nlink=2)
     
     def getxattr(self, path, name, position=0):
-        raise OSError(ENOTSUP, '')
+        raise FuseOSError(ENOTSUP)
     
     def init(self, path):
         """Called on filesystem initialization. Path is always /
@@ -545,18 +600,18 @@ class Operations(object):
         pass
     
     def link(self, target, source):
-        raise OSError(EROFS, '')
+        raise FuseOSError(EROFS)
     
     def listxattr(self, path):
         return []
-        
+    
     lock = None
     
     def mkdir(self, path, mode):
-        raise OSError(EROFS, '')
+        raise FuseOSError(EROFS)
     
     def mknod(self, path, mode, dev):
-        raise OSError(EROFS, '')
+        raise FuseOSError(EROFS)
     
     def open(self, path, flags):
         """When raw_fi is False (default case), open should return a numerical
@@ -572,7 +627,7 @@ class Operations(object):
     
     def read(self, path, size, offset, fh):
         """Returns a string containing the data requested."""
-        raise OSError(ENOENT, '')
+        raise FuseOSError(EIO)
     
     def readdir(self, path, fh):
         """Can return either a list of names, or a list of (name, attrs, offset)
@@ -580,7 +635,7 @@ class Operations(object):
         return ['.', '..']
     
     def readlink(self, path):
-        raise OSError(ENOENT, '')
+        raise FuseOSError(ENOENT)
     
     def release(self, path, fh):
         return 0
@@ -589,16 +644,16 @@ class Operations(object):
         return 0
     
     def removexattr(self, path, name):
-        raise OSError(ENOTSUP, '')
+        raise FuseOSError(ENOTSUP)
     
     def rename(self, old, new):
-        raise OSError(EROFS, '')
+        raise FuseOSError(EROFS)
     
     def rmdir(self, path):
-        raise OSError(EROFS, '')
+        raise FuseOSError(EROFS)
     
     def setxattr(self, path, name, value, options, position=0):
-        raise OSError(ENOTSUP, '')
+        raise FuseOSError(ENOTSUP)
     
     def statfs(self, path):
         """Returns a dictionary with keys identical to the statvfs C structure
@@ -607,26 +662,26 @@ class Operations(object):
         return {}
     
     def symlink(self, target, source):
-        raise OSError(EROFS, '')
+        raise FuseOSError(EROFS)
     
     def truncate(self, path, length, fh=None):
-        raise OSError(EROFS, '')
+        raise FuseOSError(EROFS)
     
     def unlink(self, path):
-        raise OSError(EROFS, '')
+        raise FuseOSError(EROFS)
     
     def utimens(self, path, times=None):
         """Times is a (atime, mtime) tuple. If None use current time."""
         return 0
     
     def write(self, path, data, offset, fh):
-        raise OSError(EROFS, '')
+        raise FuseOSError(EROFS)
 
 
 class LoggingMixIn:
     def __call__(self, op, path, *args):
         print '->', op, path, repr(args)
-        ret = '[Unknown Error]'
+        ret = '[Unhandled Exception]'
         try:
             ret = getattr(self, op)(path, *args)
             return ret
